import copy
import datetime
import logging.config

RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"
ITALIC_SEQ = "\033[3m"
UNDERLINE_SEQ = "\033[4m"

BLACK, RED, GREEN, YELLOW, ORANGE, PURPLE, CYAN, GREY = range(8)

COLORS = {
    "WARNING": YELLOW,
    "INFO": GREY,
    "DEBUG": ORANGE,
    "CRITICAL": YELLOW,
    "ERROR": RED,
}


class MillisecondFormatter(logging.Formatter):
    """A formatter for standard library 'logging' that supports '%f' wildcard in format strings."""

    converter = datetime.datetime.fromtimestamp

    def formatTime(self, record, datefmt=None):
        converter = self.converter(record.created)
        if datefmt:
            s = converter.strftime(datefmt)[:-3]
        else:
            t = converter.strftime("%Y-%m-%d %H:%M:%S")
            s = "%s,%03d" % (t, record.msecs)
        return s


class ColoredFormatter(logging.Formatter):
    def format(self, record):
        new_record = copy.copy(record)
        levelname = new_record.levelname

        if levelname in COLORS:
            levelname_color = (
                COLOR_SEQ % (30 + COLORS[levelname]) + levelname + RESET_SEQ
            )
            new_record.levelname = levelname_color
        return logging.Formatter.format(self, new_record)


def formatter_message(message, use_color=True):
    if use_color:
        message = (
            message.replace("$RESET", RESET_SEQ)
            .replace("$BOLD", BOLD_SEQ)
            .replace("$ITALIC", ITALIC_SEQ)
            .replace("$UNDER", UNDERLINE_SEQ)
        )
    else:
        message = (
            message.replace("$RESET", "")
            .replace("$BOLD", "")
            .replace("$ITALIC", "")
            .replace("$UNDER", "")
        )
    return message


def setup_logging(debug: bool = False):
    base_format = (
        "[$BOLD%(levelname)-8s$RESET] ($BOLD%(filename)s:%(lineno)d$RESET) %(message)s"
    )
    debug_base_format = (
        "[$BOLD%(levelname)-8s$RESET] {$ITALIC%(threadName)-10s$RESET} (%(funcName)s @ "
        "$BOLD%(filename)s:%(lineno)d$RESET) %(message)s "
    )
    chosen_format = debug_base_format if debug else base_format

    colored_format = formatter_message(chosen_format)
    # file_format = formatter_message(chosen_format, False)

    config = {
        "version": 1,
        "formatters": {
            "coloredFormatter": {
                "format": colored_format,
                "style": "%",
                "validate": False,
                "class": "downmixer.log.ColoredFormatter",
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": "DEBUG",
                "formatter": "coloredFormatter",
                "stream": "ext://sys.stdout",
            }
        },
        "loggers": {
            "downmixer": {
                "handlers": ["console"],
                "level": "DEBUG" if debug else "INFO",
                "propagate": True,
            }
        },
        "disable_existing_loggers": True,
    }

    logging.config.dictConfig(config)

    print("Logging setup finished")



from pathlib import Path

import mkdocs_gen_files

src_root = Path("src/downmixer")


def _create_pages_file(module_path: Path):
    with mkdocs_gen_files.open(module_path.joinpath(".pages"), "w") as f:
        print(f"title: {module_path.stem}", file=f)


def _iterate_subdirectories(dir_path: Path):
    _create_pages_file(Path("reference", dir_path.relative_to(src_root)))

    for directory in dir_path.iterdir():
        if directory.name.startswith("__") or directory.is_file():
            continue
        _iterate_subdirectories(directory)
        _create_pages_file(Path("reference", dir_path.relative_to(src_root)))


for path in src_root.glob("*"):
    if path.is_file() or path.name.startswith("__"):
        continue

    _iterate_subdirectories(path)

for path in src_root.glob("**/*.py"):
    relative_path = path.relative_to(src_root)

    is_init = relative_path.name.startswith("__")
    is_root_init = is_init and len(relative_path.parts) <= 1
    if relative_path.name == "__pycache__" or is_root_init:
        continue

    if is_init:
        doc_path = Path("reference", relative_path.parent, "index.md")
    else:
        doc_path = Path("reference", relative_path).with_suffix(".md")

    with mkdocs_gen_files.open(doc_path, "w") as f:
        if is_init:
            ident = path.with_suffix("").parent.parts[1:]
        else:
            ident = path.with_suffix("").parts[1:]

        print(f"# `{ident[-1]}`", file=f)
        print("::: " + ".".join(ident), file=f)
        if is_init:
            print("\toptions:", file=f)
            print("\t\tshow_submodules: false", file=f)



def remap(x: float, o_min: float, o_max: float, n_min: float, n_max: float) -> float:
    """Linearly map one range to another. For example, if the original range is 0 to 10 and the new range is 0 to 5,
    and `x` value of 5 will result in an output of 2.5.

    This function can handle negative values and inverted ranges. If the input is -10 to 0 and the new range is 5 to
    10, the output will still be valid.

    Args:
        x (float): The value inside the old range to be remapped.
        o_min (float): Minimum value of the old range.
        o_max (float): Maximum value of the old range.
        n_min (float): Minimum value of the new range.
        n_max (float): Maximum value of the new range.

    Returns:
        result (float): `x` modified to be fit inside the new range.
    """
    # check reversed input range
    reverse_input = False
    old_min = min(o_min, o_max)
    old_max = max(o_min, o_max)
    if not old_min == o_min:
        reverse_input = True

    # check reversed output range
    reverse_output = False
    new_min = min(n_min, n_max)
    new_max = max(n_min, n_max)
    if not new_min == n_min:
        reverse_output = True

    portion = (x - old_min) * (new_max - new_min) / (old_max - old_min)
    if reverse_input:
        portion = (old_max - x) * (new_max - new_min) / (old_max - old_min)

    result = portion + new_min
    if reverse_output:
        result = new_max - portion

    return result


def ease(x: float) -> float:
    """Returns `y` according to [this graph](https://www.desmos.com/calculator/3guvoyxg4z).

    Args:
        x (float): The value to be placed in the curve.

    Returns:
        y (float): The value placed in the curve.
    """
    return 1 - (4 * x * x * x) if x < 0.5 else ((-2 * x + 2) ** 3) / 2



"""Classes and methods to easily compare the compatibility of a result with a song being matched. Uses fuzzy string
comparison with the [RapidFuzz package](https://github.com/maxbachmann/RapidFuzz).

Matching is done individually on song name, primary artist, other artists, album name, and length - artist matches are
calculated down to a single score value (scores go from 0 to 100). Therefore, the sum can be a range of 0 to 400.
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Tuple, Optional

from rapidfuzz import fuzz

import downmixer.matching.utils
from downmixer.library import Artist, Song


class MatchQuality(Enum):
    """Thresholds to consider when getting the quality of a match. Values are based on the sum of all matches - if
    all are perfect, equals to 400.

    ## Perfect (score equals 400)
    Both songs are exactly the same.

    ## Great (score over 390)
    Extremely likely songs are the same. Different platforms usually have small discrepancies in the matching value.

    ## Good (score over 280)
    Likely a different version of the same song, like a live version for example.

    ## Mediocre (score over 150)
    Probably a cover from another artist or something else from the same artist.

    ## Bad (score equals 0)
    Not the same song.
    """

    PERFECT = 400
    GREAT = 390
    GOOD = 280
    MEDIOCRE = 150
    BAD = 0


@dataclass
class MatchResult:
    """Holds match results and provides convenient property methods to get/calculate quality and match score."""

    method: str
    name_match: float
    artists_match: list[Tuple[Artist, float]]
    album_match: float
    length_match: float

    @property
    def quality(self) -> MatchQuality:
        """Returns the match quality from the enum `MatchQuality`."""
        result = MatchQuality.PERFECT
        previous = MatchQuality.PERFECT
        for q in MatchQuality:
            if q.value <= self.sum < previous.value:
                result = q
            previous = q

        return result

    @property
    def artists_match_avg(self) -> float:
        """Averages the match score of the list of artists. Returns zero if list is empty."""
        match_values = [x[1] for x in self.artists_match]
        if len(match_values) == 0:
            return 0.0
        else:
            return sum(match_values) / len(self.artists_match)

    @property
    def sum(self) -> float:
        """Sums all matches (uses average artist match value). Maximum value is 400."""
        return (
            self.name_match
            + self.artists_match_avg
            + self.album_match
            + self.length_match
        )

    def all_above_threshold(self, threshold: float) -> bool:
        """Checks if all the scores are above the threshold value given.

        Args:
            threshold (float): Tha value that will be compared to all the values.

        Returns:
            True if every match score is higher than the threshold, false otherwise.
        """
        name_test = self.name_match >= threshold
        artists_test = self.artists_match_avg >= threshold
        album_test = self.album_match >= threshold
        length_test = self.length_match >= threshold

        return name_test and artists_test and album_test and length_test


def match(original_song: Song, result_song: Song) -> MatchResult:
    """Returns match values using RapidFuzz comparing the two given song objects.

    Args:
        original_song (Song): Song to be compared to. Should be slugified for better results.
        result_song (Song): Song being compared. Should be slugified for better results.

    Returns:
        MatchResult: Match scores of the comparison between original and result song.
    """
    song_slug = original_song.slug()
    result_slug = result_song.slug()

    name_match = _match_simple(song_slug.name, result_slug.name)
    artists_matches = _match_artist_list(song_slug, result_slug)
    if result_slug.album is not None:
        album_match = _match_simple(song_slug.album.name, result_slug.album.name)
    else:
        album_match = 0.0
    length_match = _match_length(original_song.duration, result_song.duration)

    return MatchResult(
        method="WRatio",
        name_match=name_match,
        artists_match=artists_matches,
        album_match=album_match,
        length_match=length_match,
    )


def _match_simple(str1: str, str2: str | None) -> float:
    """Calculates match score for two strings. The second string can be None and will be treated as empty if such."""
    try:
        result = fuzz.WRatio(str1, str2 if str2 is not None else "")
        match_value = result
    except ValueError:
        match_value = 0.0
    return match_value


def _match_artist_list(
    slug_song: Song, slug_result: Song
) -> list[Tuple[Artist, float]]:
    """Uses _match_simple to calculate match score of all the artists from a song."""
    artist_matches = []
    for artist in slug_song.artists:
        highest_ratio: Tuple[Optional[Artist], float] = (None, -1.0)
        for result_artist in slug_result.artists:
            ratio = _match_simple(artist.name, result_artist.name)
            if ratio > highest_ratio[1]:
                highest_ratio = (artist, ratio)

        if highest_ratio[0] is not None:
            artist_matches.append(highest_ratio)

    return artist_matches


def _match_length(len1: float, len2: float, ceiling: int = 120):
    """Plots the difference between `len1` and `len2` in a [graph](https://www.desmos.com/calculator/3guvoyxg4z) and
    returns the y value of this graph. The `ceiling` parameter defines the scale of the x-axis.
    """
    x = downmixer.matching.utils.remap(abs(len1 - len2), 0, ceiling, 0, 1)
    y = downmixer.matching.utils.ease(x) * 100
    return max(min(100, int(y)), 0)



"""Takes a playlist or song and processes it using audio and lyric providers."""

from __future__ import annotations

import asyncio
import logging
import shutil
from pathlib import Path
from typing import Type

from downmixer.file_tools import tag, utils
from downmixer.file_tools.convert import Converter
from downmixer.providers import (
    Download,
    BaseInfoProvider,
    BaseAudioProvider,
    BaseLyricsProvider,
)

logger = logging.getLogger("downmixer").getChild(__name__)


async def _convert_download(download: Download) -> Download:
    converter = Converter(download)
    return await converter.convert()


class BasicProcessor:
    def __init__(
        self,
        info_provider: BaseInfoProvider,
        audio_provider_class: Type[BaseAudioProvider],
        audio_provider_settings: str | None,
        lyrics_provider: BaseLyricsProvider,
        output_folder: Path,
        temp_folder: Path,
        threads: int = 3,
        max_retries: int = 10,
    ):
        """Basic processing class to search an ID and download it, using the providers passed on by the user. For
        playlist downloads, it uses an [`asyncio.Semaphore`](
        https://docs.python.org/3/library/asyncio-sync.html#semaphore) with a value of `threads` - so the number of
        concurrent downloads is equal to the `threads` value.

        Args:
            info_provider (BaseInfoProvider): Class instance to use when searching an ID.
            audio_provider_class (Type[BaseAudioProvider]): Class reference to use when downloading songs.
            audio_provider_settings (str, optional): JSON formatted settings for the BaseAudioProvider.
            lyrics_provider (BaseLyricsProvider): Class instance to use when downloading lyrics.
            output_folder (str): Folder path where the final file will be placed.
            temp_folder (str): Folder path where temporary files will be placed and removed from when processing
                is finished.
            threads (int): Amount of threads that will simultaneously process songs.
        """
        self.output_folder: Path = Path(output_folder).absolute()
        self.temp_folder = temp_folder

        self.info_provider = info_provider
        self.audio_provider_class = audio_provider_class
        self.audio_provider_settings = audio_provider_settings
        self.lyrics_provider = lyrics_provider

        self.max_retries = max_retries
        self.semaphore = asyncio.Semaphore(threads)

    async def _get_lyrics(self, download: Download):
        # TODO: Test if lyrics are actually working
        lyrics_results = await self.lyrics_provider.search(download.song)
        if lyrics_results is not None:
            lyrics = await self.lyrics_provider.get_lyrics(lyrics_results[0])
            download.song.lyrics = lyrics

    async def pool_processing(self, song_id: str):
        async with self.semaphore:
            logger.debug(f"Processing song '{song_id}'")
            retries = 0
            while retries <= self.max_retries:
                try:
                    await self.process_song(song_id)
                    return
                except Exception as e:
                    # TODO: Pick out exceptions instead of catching all exceptions
                    logger.warning(
                        f"Error processing song '{song_id}', retrying ({self.max_retries - retries} left)",
                        exc_info=e,
                    )
                    retries += 1

            logger.error(f"Max retries exceeded for song '{song_id}'")

    async def process_playlist(self, playlist_id: str):
        """Searches and downloads all songs in a playlist using a queue with limited threads.

        Args:
            playlist_id (str): ID for the playlist to be downloaded."""
        songs = self.info_provider.get_all_playlist_songs(playlist_id)

        tasks = [self.pool_processing(s.id) for s in songs]
        await asyncio.gather(*tasks)

    async def process_song(self, song_id: str):
        """Searches and downloads a single song based on data provided by a `BaseInfoProvider`.

        Args:
            song_id (str): Valid ID of a single track.
        """
        song = self.info_provider.get_song(song_id)
        audio_provider = self.audio_provider_class(self.audio_provider_settings)

        result = await audio_provider.search(song)
        if result is None:
            logger.warning("Song not found", extra={"songinfo": song.__dict__})
            return
        downloaded = await audio_provider.download(result[0], self.temp_folder)
        converted = await _convert_download(downloaded)

        await self._get_lyrics(converted)
        tag.tag_download(converted)

        new_name = (
            utils.make_sane_filename(converted.song.title) + converted.filename.suffix
        )

        self.output_folder.mkdir(parents=True, exist_ok=True)
        logger.debug(
            f"Moving file from '{converted.filename}' to '{self.output_folder}'"
        )
        shutil.move(converted.filename, self.output_folder.joinpath(new_name))



"""Defines base provider classes and give default lyrics and audio providers."""

from __future__ import annotations

import importlib
import pkgutil
import sys
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Optional, Type

from downmixer.file_tools import AudioCodecs
from downmixer.library import Song, Playlist
from downmixer.matching import MatchResult, MatchQuality


class ResourceType(Enum):
    SONG = 1
    ALBUM = 2
    ARTIST = 3
    PLAYLIST = 4
    USER = 5


@dataclass
class AudioSearchResult:
    """Holds data about a result from a `BaseAudioProvider` instance."""

    provider: str
    match: MatchResult
    download_url: str
    _original_song: Song
    _result_song: Song

    @property
    def song(self) -> Song:
        """Compares the match quality with a set threshold and returns the most appropriate choice between the
        original song from Spotify or the result given by the provider.

        Returns:
            song (Song): The appropriate song object.
        """
        if self.match.quality == MatchQuality.PERFECT or self.match.quality.GREAT:
            return self._original_song
        else:
            return self._result_song


@dataclass
class LyricsSearchResult:
    """Holds data about a result from a `BaseLyricsProvider` instance."""

    provider: str
    match: MatchResult
    name: str
    artist: str
    url: str


@dataclass
class Download(AudioSearchResult):
    """A child of `AudioSearchResult` which has been successfully downloaded. Contains data about the downloaded
    file and its path.

    Attributes:
        filename (Path): Path to the downloaded song on the system.
        bitrate (float): The file's bitrate in kbps.
        audio_codec (AudioCodecs): One of the supported audio codecs from `AudioCodecs` enum.
    """

    filename: Path
    bitrate: float
    audio_codec: AudioCodecs

    @classmethod
    def from_parent(
        cls,
        parent: AudioSearchResult,
        filename: Path,
        bitrate: float,
        audio_codec: AudioCodecs,
    ):
        """Make a Download instance with the information from a parent `ProviderSearchResult` class.

        Args:
            parent (AudioSearchResult): The class instance being used.
            filename (Path): Path to the downloaded song on the system.
            bitrate (float): The file's bitrate in kbps.
            audio_codec (AudioCodecs): One of the supported audio codecs from `AudioCodecs` enum.

        Returns:
            cls (Download): Download instance with attributes from the `parent` object and other provided info.
        """
        return cls(
            provider=parent.provider,
            match=parent.match,
            _result_song=parent._result_song,
            _original_song=parent._original_song,
            download_url=parent.download_url,
            filename=filename,
            bitrate=bitrate,
            audio_codec=audio_codec,
        )


class BaseAudioProvider:
    """
    Base class for all audio providers. Defines the interface that any audio provider in Downmixer should use.
    """

    provider_name = ""

    def __init__(self, options: dict = None):
        """Initializes the provider.

        Args:
            options (dict): Dictionary of options to pass to the provider. See documentation for each provider for
                available options.
        """
        self.options = options

    async def search(self, song: Song) -> Optional[list[AudioSearchResult]]:
        """Retrieves search results as list of `AudioSearchResult` objects ordered by match, highest to lowest.
        Can return None if a problem occurs.

        Args:
            song (Song): Song object which will be searched.

        Returns:
            Optional list containing the search results as `AudioSearchResult` objects.
        """
        raise NotImplementedError

    async def download(
        self, result: AudioSearchResult, path: Path
    ) -> Optional[Download]:
        """Downloads, using this provider, a search result to the path specified.

        Args:
            result (AudioSearchResult): The `AudioSearchResult` that matches with this provider class.
            path (Path): The folder (not filename) in which the file will be downloaded.

        Returns:
            Download object with the downloaded file information.
        """
        raise NotImplementedError


class BaseLyricsProvider:
    """
    Base class for all lyrics providers. Defines the interface that any lyrics provider in Downmixer should use.
    """

    provider_name = ""

    def __init__(self, options: dict = None):
        """Initializes the provider.

        Args:
            options (dict): Dictionary of options to pass to the provider. See documentation for each provider for
                available options.
        """
        self.options = options

    async def search(self, song: Song) -> Optional[list[LyricsSearchResult]]:
        """Retrieves search results as list of `LyricsSearchResult` objects ordered by match, highest to lowest.
        Can return None if a problem occurs.

        Args:
            song (Song): Song object which will be searched.

        Returns:
            Optional list containing the search results as `LyricsSearchResult` objects.
        """
        raise NotImplementedError

    async def get_lyrics(self, result: LyricsSearchResult) -> Optional[str]:
        """Retrieves lyrics for a specific search result from this provider.

        Args:
            result (LyricsSearchResult): The song being searched.

        Returns:
            Optional string with the lyrics of the song.
        """
        raise NotImplementedError


class NotConnectedException(Exception):
    pass


# noinspection PyTypeChecker
class BaseInfoProvider:
    """Base class for all song info providers. Defines the interface that any song info provider in Downmixer should
    use.

    Info providers are streaming platforms where a user has a library of songs, a.k.a. Spotify, YouTube Music,
    Deezer, Apple Music, etc. Used to get the user's library and read the data of that song on said platform.
    """

    connected = False

    def __init__(self, options: dict = None):
        """Initializes the provider.

        Args:
            options (dict): Dictionary of options to pass to the provider. See documentation for each provider for
                available options.
        """
        self.options = options

    def check_valid_url(self, url: str, type_filter: list[ResourceType] = None) -> bool:
        pass

    def get_resource_type(self, value: str) -> ResourceType | None:
        """Determines the `ResourceType` of the library item.

        Returns:
            Instance of `ResourceType` enum.
        """
        pass

    def connect(self):
        """"""
        pass

    def get_song(self, track_id: str) -> Song:
        """Retrieve a song from the info provider. Returns a new Song object with the metadata from
        the provider.

        Args:
            track_id (str): A string containing a valid ID for the provider.

        Returns:
            Song object with the metadata retrieved from the provider.
        """
        if not self.connected:
            raise NotConnectedException(
                "Not connected to platform, cannot retrieve data"
            )

        if not self.check_valid_url(track_id, [ResourceType.SONG]):
            raise ValueError(f"{track_id} is an invalid song URL or ID")

        pass

    def get_all_playlist_songs(self, playlist_id: str) -> list[Song]:
        """Retrieves the all the songs from a playlist as a list.
        Args:
            playlist_id (str): A string containing a valid ID for the provider.

        Returns:
            User's playlists as a list of Song objects.
        """
        if not self.connected:
            raise NotConnectedException(
                "Not connected to platform, cannot retrieve data"
            )

        if not self.check_valid_url(
            playlist_id, [ResourceType.PLAYLIST, ResourceType.ALBUM]
        ):
            raise ValueError(f"{playlist_id} is an invalid playlist URL or ID")

        pass

    def get_all_user_playlists(self) -> list[Playlist]:
        """Retrieves the all the user's playlists in a list.

        Returns:
            User's playlists as a list of Playlist objects.
        """
        if not self.connected:
            raise NotConnectedException(
                "Not connected to platform, cannot retrieve data"
            )
        pass

    def get_all_user_albums(self) -> list[Playlist]:
        """Retrieves the all the user's saved albums in a list.

        Returns:
            User's albums as a list of Album objects.
        """
        if not self.connected:
            raise NotConnectedException(
                "Not connected to platform, cannot retrieve data"
            )
        pass

    def get_all_user_songs(self) -> list[Song]:
        """Retrieves the all the user's liked/saved songs in a list (for example, on Spotify, should return user's
        saved tracks).

        Returns:
            User's playlists as a list of Playlist objects.
        """
        if not self.connected:
            raise NotConnectedException(
                "Not connected to platform, cannot retrieve data"
            )
        pass


# TODO: Add support for importing third-party providers
def _import_submodules(prefix: str):
    providers_path = str(Path(__file__).parent.joinpath(prefix).absolute())
    current_package_name = sys.modules[__name__].__name__

    for loader, name, is_pkg in pkgutil.walk_packages([providers_path]):
        importlib.import_module(f"{current_package_name}.{prefix}.{name}")


def get_all_audio_providers() -> list[Type[BaseAudioProvider]]:
    """Imports providers from internal submodules folder and returns a list containing all subclasses of
    BaseAudioProvider."""
    _import_submodules("audio")
    return BaseAudioProvider.__subclasses__()


def get_all_lyrics_providers() -> list[Type[BaseLyricsProvider]]:
    """Imports providers from internal submodules folder and returns a list containing all subclasses of
    BaseLyricsProvider."""
    _import_submodules("lyrics")
    return BaseLyricsProvider.__subclasses__()


def get_all_info_providers() -> list[Type[BaseInfoProvider]]:
    """Imports providers from internal submodules folder and returns a list containing all subclasses of
    BaseInfoProvider."""
    _import_submodules("info")
    return BaseInfoProvider.__subclasses__()



import argparse
import asyncio
import json
import logging
import os
import tempfile
import time
from pathlib import Path

from downmixer import processing, log
from downmixer import providers
from downmixer.providers import ResourceType

logger = logging.getLogger("downmixer").getChild(__name__)

parser = argparse.ArgumentParser(
    prog="downmixer", description="Easily sync tracks from any streaming service."
)
parser.add_argument("procedure", choices=["download"])
parser.add_argument(
    "id",
    help="A valid Spotify ID, URI or URL for a track, album or playlist.",
)
parser.add_argument(
    "-t",
    "--threads",
    default=3,
    type=int,
    help="Number of threads to use for parallel downloads.",
)
parser.add_argument(
    "-o",
    "--output-folder",
    type=Path,
    default=os.curdir,
    dest="output",
    help="Path to the folder in which the final processed files will be placed.",
)
parser.add_argument(
    "-ip",
    "--info-provider",
    type=str,
    default="SpotifyInfoProvider",
    choices=[x.__name__ for x in providers.get_all_info_providers()],
    help=f"Info provider extending BaseInfoProvider to use. Defaults to 'SpotifyInfoProvider'.",
)
parser.add_argument(
    "-ip-settings",
    "--info-provider-settings",
    type=str,
    default=None,
    help="Settings for the info provider as a JSON string. See documentation for available options for each provider.",
)
parser.add_argument(
    "-ap",
    "--audio-provider",
    type=str,
    default="YouTubeMusicAudioProvider",
    choices=[x.__name__ for x in providers.get_all_audio_providers()],
    help=f"Audio provider extending BaseAudioProvider to use. Defaults to 'YouTubeMusicAudioProvider'.",
)
parser.add_argument(
    "-ap-settings",
    "--audio-provider-settings",
    type=str,
    default=None,
    help="Settings for the audio provider as a JSON string. See documentation for available options for each provider.",
)
parser.add_argument(
    "-lp",
    "--lyrics-provider",
    type=str,
    default="AZLyricsProvider",
    choices=[x.__name__ for x in providers.get_all_lyrics_providers()],
    help=f"Lyrics provider extending BaseLyricsProvider to use. Defaults to 'AZLyricsProvider'.",
)
parser.add_argument(
    "-lp-settings",
    "--lyrics-provider-settings",
    type=str,
    default=None,
    help="Settings for the lyrics provider as a JSON string. See documentation for available options for each provider.",
)
args = parser.parse_args()


def command_line():
    log.setup_logging(debug=True)

    if args.procedure == "download":
        logger.info("Running download command")

        with tempfile.TemporaryDirectory() as temp:
            logger.debug(f"temp folder: {temp}")

            selected_info_provider = [
                x
                for x in providers.get_all_info_providers()
                if x.__name__ == args.info_provider
            ][0]
            selected_audio_provider = [
                x
                for x in providers.get_all_audio_providers()
                if x.__name__ == args.audio_provider
            ][0]
            selected_lyrics_provider = [
                x
                for x in providers.get_all_lyrics_providers()
                if x.__name__ == args.lyrics_provider
            ][0]

            ip_settings = (
                json.loads(args.info_provider_settings)
                if args.info_provider_settings
                else None
            )
            ap_settings = (
                json.loads(args.audio_provider_settings)
                if args.audio_provider_settings
                else None
            )
            lp_settings = (
                json.loads(args.lyrics_provider_settings)
                if args.lyrics_provider_settings
                else None
            )

            processor = processing.BasicProcessor(
                selected_info_provider(ip_settings),
                selected_audio_provider,
                ap_settings,
                selected_lyrics_provider(lp_settings),
                args.output,
                Path(temp),
                args.threads,
            )

            logger.debug(
                f"Initialized processor with providers: {processor.info_provider.__class__.__name__}, "
                f"{selected_audio_provider.__class__.__name__}, "
                f"{processor.lyrics_provider.__class__.__name__}"
            )

            if not processor.info_provider.check_valid_url(
                args.id,
                [ResourceType.SONG, ResourceType.PLAYLIST, ResourceType.ALBUM],
            ):
                raise ValueError("id provided isn't valid")
            rtype = processor.info_provider.get_resource_type(args.id)

            start = time.time()
            if rtype == ResourceType.SONG:
                logger.info("Downloading one track")
                asyncio.run(processor.process_song(args.id))
            else:
                logger.info("Downloading many tracks")
                loop = asyncio.new_event_loop()
                loop.run_until_complete(processor.process_playlist(args.id))
                loop.close()

            logger.info(f"Finished processing in {time.time() - start} seconds")

    exit()


if __name__ == "__main__":
    command_line()



from __future__ import annotations


def merge_dicts_with_priority(dict1: dict, dict2: dict | None) -> dict:
    """Merges two dictionaries with priority to `dict1`.

    All keys and values from `dict1` will be returned in the new dictionary, substituting keys with different
    values from `dict2`. Keys and values exclusive to `dict2` will be returned as well. The function is
    recursively called for all nested dictionaries.

    Args:
        dict1 (dict): Priority dictionary to merge. Values in dict1 are always in the returned dictionary.
        dict2 (dict): Secondary dictionary to merge.

    Returns:
        dict: Merged dictionary.
    """
    new_dict = dict1.copy()

    if dict2 is not None:
        for key, value in dict2.items():
            if (
                key in new_dict
                and isinstance(new_dict[key], dict)
                and isinstance(value, dict)
            ):
                new_dict[key] = merge_dicts_with_priority(new_dict[key], value)
            elif key not in new_dict:
                new_dict[key] = value

    return new_dict



from typing import Optional

import requests
from bs4 import BeautifulSoup, Comment, ResultSet

from downmixer import matching
from downmixer.library import Song, Artist
from downmixer.providers import BaseLyricsProvider, LyricsSearchResult

COPYRIGHT_DISCLAIMER = (
    "Usage of azlyrics.com content by any third-party lyrics provider is prohibited by our "
    "licensing agreement. Sorry about that."
)

# TODO: Remove AZLyrics and add Genius provider


def search_result_from_azlyrics(
    result: ResultSet, original_song: Song
) -> LyricsSearchResult:
    """Create a LyricsSearchResult instance from a
    [Beautiful Soup 4 `ResultSet`](https://www.crummy.com/software/BeautifulSoup/bs4/doc/#find-all) from AZLyrics.

    Args:
        result (bs4.ResultSet):
        original_song (Song): Instance of a song from Spotify that will be compared against this search result.

    Returns:
        LyricsSearchResult from AZLyrics.
    """
    strings = result[0].find_all("b")
    name = strings[0].text[1:-1]
    artist = strings[1].text

    result_song = Song(name=name, artists=[Artist(name=artist)])
    return LyricsSearchResult(
        provider="azlyrics",
        match=matching.match(original_song, result_song),
        name=name,
        artist=artist,
        url=result[0]["href"],
    )


class AZLyricsProvider(BaseLyricsProvider):
    provider_name = "azlyrics"

    def __init__(self, options: dict = None):
        super().__init__({})
        headers = {
            "Connection": "keep-alive",
            "Pragma": "no-cache",
            "Cache-Control": "no-cache",
            "sec-ch-ua": '"Chromium";v="104", " Not A;Brand";v="99", "Google Chrome";v="104"',
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
            "(KHTML, like Gecko) Chrome/104.0.0.0 Safari/537.36",
            "Accept": "*/*",
            "Sec-Fetch-Site": "same-origin",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Dest": "empty",
            "Accept-Language": "en-US;q=0.8,en;q=0.7",
        }

        self.session = requests.Session()
        self.session.headers.update(headers)

        resp = self.session.get("https://www.azlyrics.com/geo.js")

        js_code = resp.text
        start_index = js_code.find('value"') + 9
        end_index = js_code[start_index:].find('");')

        self.x_code = js_code[start_index : start_index + end_index]

    async def get_lyrics(self, search_result: LyricsSearchResult) -> Optional[str]:
        response = self.session.get(search_result.url)
        soup = BeautifulSoup(response.content, "html.parser")

        div_tags = soup.find_all("div", class_=False, id_=False)
        for d in div_tags:
            comments = d.find_all(string=lambda x: isinstance(x, Comment))
            # All AZLyrics lyrics pages have this disclaimer as an HTML comment on the top of the div
            # containing the lyrics. We can use it to identify which div has lyrics.
            if len(comments) != 0 and COPYRIGHT_DISCLAIMER in comments[0]:
                lyrics_div = d
                lyrics = lyrics_div.get_text().strip()
                return lyrics

        return None

    async def search(self, song: Song) -> Optional[list[LyricsSearchResult]]:
        params = {"q": song.full_title, "x": self.x_code, "w": "songs"}

        response = self.session.get(
            "https://search.azlyrics.com/search.php", params=params
        )
        soup = BeautifulSoup(response.content, "html.parser")

        td_tags = soup.find_all("td")
        if len(td_tags) == 0:
            return None

        results: list[LyricsSearchResult] = []
        result_list = [x.find_all("a", href=True) for x in td_tags]
        for r in result_list:
            if len(r) == 0:
                continue
            # The first and last td tags in AZLyrics have the page buttons. Skip those when we meet them.
            elif r[0].has_attr("class") and "btn" in r[0]["class"]:
                continue
            else:
                results.append(search_result_from_azlyrics(r, song))

        ordered_results = sorted(results, reverse=True, key=lambda x: x.match.sum)

        return ordered_results


def instance():
    return AZLyricsProvider



from __future__ import annotations

import logging
import re

import spotipy

from downmixer import utils
from downmixer.library import Playlist
from downmixer.providers import BaseInfoProvider, ResourceType
from .library import SpotifySong, SpotifyPlaylist, SpotifyAlbum

logger = logging.getLogger("downmixer").getChild(__name__)

resource_type_map = {
    ResourceType.SONG: "track",
    ResourceType.ALBUM: "album",
    ResourceType.PLAYLIST: "playlist",
    ResourceType.ARTIST: "artist",
}


def _get_all(func, limit=50, *args, **kwargs):
    counter = 0
    next_url = ""
    items = []

    while next_url is not None:
        results = func(*args, **kwargs, limit=limit, offset=limit * counter)
        next_url = results["next"]
        counter += 1
        items += results["items"]

    return items


class SpotifyInfoProvider(BaseInfoProvider):
    def __init__(self, options: dict = None):
        default_options = {
            "auth": {
                "scope": "user-library-read,user-follow-read,playlist-read-private",
            }
        }
        options = utils.merge_dicts_with_priority(default_options, options)
        super().__init__(options)

        # TODO: Manage auth properly
        self.client = spotipy.Spotify(
            auth_manager=spotipy.SpotifyOAuth(**options["auth"])
        )

        self.connected = True

    def get_resource_type(self, value: str) -> ResourceType | None:
        if not self.check_valid_url(value):
            return None

        pattern = r"spotify(?:.com)?(?::|\/)(\w*)(?::|\/)(?:\w{20,24})"
        matches = re.search(pattern, value)

        if matches is None:
            return None
        else:
            return list(resource_type_map.keys())[
                list(resource_type_map.values()).index(matches.group(1).lower())
            ]

    def check_valid_url(self, url: str, type_filter: list[ResourceType] = None) -> bool:
        if type_filter is None:
            type_filter = [e for e in ResourceType]

        for t in type_filter:
            regex = r"spotify.*" + resource_type_map[t] + r"(?::|\/)(\w{20,24})"
            if re.search(regex, url) is not None:
                return True

        return False

    def get_song(self, track_id: str) -> SpotifySong:
        super().get_song(track_id)

        result = self.client.track(track_id)
        return SpotifySong.from_provider(result)

    def get_all_playlist_songs(self, playlist_id: str) -> list[SpotifySong]:
        super().get_all_playlist_songs(playlist_id)

        if self.check_valid_url(playlist_id, [ResourceType.PLAYLIST]):
            results = _get_all(
                self.client.playlist_items, limit=50, playlist_id=playlist_id
            )
            return SpotifySong.from_provider_list(results)
        else:
            album_info = self.client.album(playlist_id)
            results = _get_all(self.client.album_tracks, limit=50, album_id=playlist_id)
            return SpotifySong.from_provider_list(
                results, extra_data={"album": album_info}
            )

    def get_all_user_playlists(self) -> list[SpotifyPlaylist]:
        super().get_all_user_playlists()

        results = _get_all(self.client.current_user_playlists)
        return SpotifyPlaylist.from_provider_list(results)

    def get_all_user_albums(self) -> list[Playlist]:
        super().get_all_user_albums()

        results = _get_all(self.client.current_user_saved_albums, limit=50)
        return SpotifyAlbum.from_provider_list(results)

    def get_all_user_songs(self) -> list[SpotifySong]:
        super().get_all_user_songs()

        results = _get_all(self.client.current_user_saved_tracks, limit=50)
        return SpotifySong.from_provider_list(results)



"""Data classes to hold standardized metadata about songs, artists, and albums."""

from dataclasses import dataclass
from enum import Enum
from typing import Optional, Any

from slugify import slugify


class AlbumType(Enum):
    ALBUM = 1
    SINGLE = 2
    COMPILATION = 3


class BaseLibraryItem:
    """Base class for library items containing standard methods to easily create class instances from the Spotify
    API. Child classes of this class must be
    [implemented in all info providers.](../providers/library.py file.md)
    """

    @classmethod
    def from_provider(cls, data: Any, extra_data: dict = None):
        """Create an instance of this class from data coming from a provider's API.

        Args:
            data (Any): Data from the provider's API.
            extra_data (dict, optional): Extra data from provider's API to be used to make instances of this class.

        Returns:
             An instance of this class.
        """
        pass

    @classmethod
    def from_provider_list(cls, data: list[Any], extra_data: dict = None) -> list:
        """Creates a list of instances of this class from a list of objects with data coming from a provider's API.

        Args:
            data (list[Any]): List of objects with data from the provider's API.
            extra_data (dict, optional): Extra data from provider's API to be used to make instances of this class.

        Returns:
            A list with instances of this class.
        """
        return [cls.from_provider(x) for x in data]


@dataclass
class Artist(BaseLibraryItem):
    """Holds info about an artist."""

    name: str
    images: Optional[list[str]] = None
    genres: Optional[list[str]] = None
    id: Optional[str] = None
    url: Optional[str] = None

    def slug(self) -> "Artist":
        """Returns self with sluggified text attributes."""
        return Artist(
            name=slugify(self.name),
            images=self.images,
            genres=[slugify(x) for x in self.genres] if self.genres else None,
            id=self.id,
            url=self.url,
        )


@dataclass
class Album(BaseLibraryItem):
    """Holds info about an album. `cover` should be a string containing a valid URL."""

    name: str
    available_markets: Optional[list[str]] = None
    artists: Optional[list[Artist]] = None
    date: Optional[str] = None
    track_count: Optional[int] = None
    cover: Optional[str] = None
    id: Optional[str] = None
    url: Optional[str] = None

    def slug(self) -> "Album":
        """Returns self with sluggified text attributes."""
        return Album(
            name=slugify(self.name),
            available_markets=self.available_markets,
            artists=[x.slug for x in self.artists] if self.artists else None,
            date=self.date,
            track_count=self.track_count,
            cover=self.cover,
            id=self.id,
            url=self.url,
        )


@dataclass
class Song(BaseLibraryItem):
    """Holds info about a song."""

    name: str
    artists: list[Artist]
    duration: float = 0  # in seconds
    album: Optional[Album] = None
    available_markets: Optional[list[str]] = None
    date: Optional[str] = None
    track_number: Optional[int] = None
    isrc: Optional[str] = None
    lyrics: Optional[str] = None
    id: Optional[str] = None
    url: Optional[str] = None
    cover: Optional[str] = None

    def slug(self) -> "Song":
        """Returns self with sluggified text attributes."""
        return Song(
            name=slugify(self.name),
            artists=[x.slug() for x in self.artists],
            album=self.album.slug() if self.album else None,
            available_markets=self.available_markets,
            date=self.date,
            duration=self.duration,
            track_number=self.track_number,
            isrc=self.isrc,
            id=self.id,
            url=self.url,
            lyrics=slugify(self.lyrics) if self.lyrics else None,
        )

    @property
    def title(self) -> str:
        """str: Title of the song, including artist, in the format '[primary artist] - [song name]'."""
        return self.artists[0].name + " - " + self.name

    @property
    def full_title(self) -> str:
        """str: Full title of the song, including all artists, in the format [artist 1, artist 2, ...] - [song name]."""
        return (", ".join(x.name for x in self.artists)) + " - " + self.name

    @property
    def all_artists(self) -> str:
        """str: All artists' names, separated by a comma."""
        return ", ".join(x.name for x in self.artists)


@dataclass
class Playlist(BaseLibraryItem):
    name: str
    description: Optional[str] = None
    tracks: Optional[list[Song]] = None
    images: Optional[list[dict]] = None
    id: Optional[str] = None
    url: Optional[str] = None



from __future__ import annotations

import asyncio
import functools
import json
import logging
from http.cookiejar import CookieJar
from pathlib import Path
from typing import Optional, Any, Callable

import yt_dlp
import ytmusicapi

from downmixer import matching, utils
from downmixer.file_tools import AudioCodecs
from downmixer.library import Artist, Album, Song
from downmixer.providers import BaseAudioProvider, AudioSearchResult, Download

logger = logging.getLogger("downmixer").getChild(__name__)


def artist_from_ytmusic(data: dict[str, Any]) -> Artist:
    """Create an Artist instance from a dict provided by the YouTube Music API search function. Sadly the only
    metadata of use form the search results is the title of the artist. More info on the result's schema [here](
    https://ytmusicapi.readthedocs.io/en/stable/reference.html#search).

    Args:
        data (dict):
            Dictionary provided by the YouTube Music API from the
            [`ytmusicapi`](https://github.com/sigma67/ytmusicapi) package.

    Returns:
        Artist instance populated from the dict.
    """
    return Artist(name=data["name"])


def album_from_ytmusic(data: dict[str, Any]) -> Album:
    """Create an Album instance from a dict provided by the YouTube Music API search function. Sadly the only
    metadata of use form the search results is the title of the album. More info on the result's schema [here](
    https://ytmusicapi.readthedocs.io/en/stable/reference.html#search).

    Args:
        data (dict):
            Dictionary provided by the YouTube Music API from the
            [`ytmusicapi`](https://github.com/sigma67/ytmusicapi) package.

    Returns:
        Album instance populated from the dict.
    """
    return Album(name=data["name"])


def song_from_ytmusic(data: dict[str, Any]) -> Song:
    """Create a Song instance from a dict provided by the YouTube Music API search function, including the Artist and
    Album objects. More info on the result's schema
    [here](https://ytmusicapi.readthedocs.io/en/stable/reference.html#search).

    Args:
        data (dict):
            Dictionary provided by the YouTube Music API from the
            [`ytmusicapi`](https://github.com/sigma67/ytmusicapi) package.

    Returns:
        Song instance populated from the dict.
    """
    return Song(
        name=data["title"],
        artists=[artist_from_ytmusic(x) for x in data["artists"]],
        album=album_from_ytmusic(data["album"]),
        duration=data["duration_seconds"],
        url=f"https://music.youtube.com/watch?v={data['videoId']}",
    )


def search_result_from_ytmusic(
    original_song: Song, result_song: Song
) -> AudioSearchResult:
    """Create an AudioSearchResult instance from a dict provided by the YouTube Music API search function. Sadly the only
    metadata of use form the search results is the title of the artist. More info on the result's schema [here](
    https://ytmusicapi.readthedocs.io/en/stable/reference.html#search).

    Args:
        original_song (Song): Instance of a song from Spotify that will be compared against this search result.
        result_song (Song): Song extracted from the search result info.

    Returns:
        AudioSearchResult from YT Music.
    """
    return AudioSearchResult(
        provider="ytmusic",
        _original_song=original_song,
        _result_song=result_song,
        match=matching.match(original_song, result_song),
        download_url=result_song.url,
    )


def _get_auth_headers(cookiejar: CookieJar) -> str | None:
    """Makes the headers that `ytmusicapi` expects to authorize YT Music requests from cookies extracted from the
    browser_specification. Returns JSON as a string.

    If no YouTube cookies are found, returns None.
    """
    cookie_strings = [
        x.name + "=" + x.value.replace('"', "")
        for x in cookiejar
        if "youtube" in x.domain
    ]
    cookies = ";".join(cookie_strings)

    if len(cookies) == 0:
        return None

    auth_headers = {
        "origin": "https://music.youtube.com",
        "x-origin": "https://music.youtube.com",
        "cookie": cookies,
    }
    return json.dumps(auth_headers)


async def _run_in_loop(func: Callable, kwargs: dict) -> Any:
    """Runs a synchronous function asynchronously using the current event loop.

    Args:
        func: Function to be called.
        kwargs: Dictionary of keyword arguments that will be passed to the function.

    Returns:
        (Any): Whatever the `func` function originally returns.
    """
    loop = asyncio.get_event_loop()
    return await loop.run_in_executor(
        None,
        functools.partial(func, **kwargs),
    )


class YouTubeMusicAudioProvider(BaseAudioProvider):
    provider_name = "youtube-music"

    def __init__(self, options: dict = None):
        default_options = {"encoding": "UTF-8", "format": "bestaudio"}
        options = utils.merge_dicts_with_priority(default_options, options)
        super().__init__(options)

        self.youtube_dl = yt_dlp.YoutubeDL(self.options)
        logger.debug(f"Initialized YoutubeDL client with options: {self.options}")

        auth_headers = _get_auth_headers(self.youtube_dl.cookiejar)

        # TODO: Some testing relating to this ⬇️
        # For some reason some songs like 70tjloUDVlGYkapPPTWRxU weren't found via ISRC if the language param was not
        # specified 🤷🏻‍♀️. Selecting English bought a completely fucked up result too. I copied "de" (aka German)
        # from spotDL
        self.client = ytmusicapi.YTMusic(auth=auth_headers, language="de")

    async def search(self, song: Song) -> Optional[list[AudioSearchResult]]:
        logger.info(f"Initializing search for song '{song.title}' with URI {song.id}")
        if song.isrc:
            query = song.isrc
        else:
            query = song.title

        # TODO: redo search if ISRC isn't found
        logger.debug(f"Searching query '{query}'")
        results = self.client.search(query, filter="songs", ignore_spelling=True)

        if len(results) == 0:
            logger.warning("Search returned no results")
            return None

        result_objects = []
        for r in results:
            result_song = song_from_ytmusic(r)
            search_result = search_result_from_ytmusic(song, result_song)
            result_objects.append(search_result)
            logger.debug(
                f"Found song '{result_song.title}' with URL {result_song.url}, match value {search_result.match.sum}"
            )

        ordered_results = sorted(
            result_objects, reverse=True, key=lambda x: x.match.sum
        )
        logger.debug(f"Ordered {len(ordered_results)} results")
        return ordered_results

    async def download(
        self, result: AudioSearchResult, path: Path
    ) -> Optional[Download]:
        logger.info(
            f"Starting download for search result '{result.song.title}' with URL {result.download_url}"
        )

        # Set output path of YoutubeDL on the fly
        self.youtube_dl.params["outtmpl"]["default"] = (
            str(path.absolute()) + "/%(id)s.%(ext)s"
        )
        url = result.download_url
        metadata = await _run_in_loop(
            self.youtube_dl.extract_info, {"url": url, "download": True}
        )
        logger.info("Finished downloading")

        downloaded = metadata["requested_downloads"][0]

        logger.debug("Creating download object")
        return Download.from_parent(
            parent=result,
            filename=Path(downloaded["filepath"]),
            bitrate=downloaded["abr"],
            audio_codec=AudioCodecs(downloaded["acodec"]),
        )


def instance():
    return YouTubeMusicAudioProvider



import re


def make_sane_filename(filename: str) -> str:
    """Removed common illegal characters from names, especially for Windows.

    Args:
        filename (str): Filename to be processed

    Returns:
        result (str): Filename with illegal characters replaced with '-'.
    """
    return re.sub(r"[/\\?%*:|\"<>\x7F\x00-\x1F]", "-", filename)



"""Code relating to the manipulation of files - converting and tagging audio files specifically."""

from enum import Enum


class Format(Enum):
    MP3 = "mp3"
    FLAC = "flac"
    WAV = "wav"
    OPUS = "opus"


# TODO: Deal with the fact there's like 5000 different combinations of this
# https://developer.mozilla.org/en-US/docs/Web/Media/Formats/codecs_parameter#mpeg-4_audio
class AudioCodecs(Enum):
    MP4A_40_5 = "mp4a.40.5"
    MP4A_40_2 = "mp4a.40.2"
    OPUS = "opus"



import copy
import logging
import os
from pathlib import Path

from ffmpeg.asyncio import FFmpeg

from downmixer.file_tools import Format
from downmixer.providers import Download

logger = logging.getLogger("downmixer").getChild(__name__)


class Converter:
    def __init__(
        self, download: Download, format: Format = Format.MP3, bitrate: str = "320k"
    ):
        """Holds information for FFmpeg to convert a download. By default, uses MP3 output format and 320kbps bitrate.

        Args:
            download (Download): Download object to be converted.
            format (Format): Output format from the Format enum.
            bitrate (str): Bitrate in kbps as a string denoting value with a 'k' in the end. Passed directly into FFmpeg.
        """
        self.download = download
        self.format = format
        self.bitrate = bitrate

    async def convert(self, delete_original: bool = True) -> Download:
        logger.info("Starting conversion")

        # TODO: Check if file already exists
        output = str(self.download.filename).replace(
            self.download.filename.suffix, "." + self.format.value
        )
        ffmpeg = (
            FFmpeg()
            .option("vn")
            .input(str(self.download.filename))
            .output(output, {"b:a": self.bitrate})
        )

        @ffmpeg.on("start")
        def on_start(arguments):
            logger.debug("Arguments: " + ", ".join(arguments))

        @ffmpeg.on("stderr")
        def on_stderr(line):
            logger.debug(line)

        @ffmpeg.on("progress")
        def on_progress(progress):
            logger.info(progress)

        @ffmpeg.on("terminated")
        def on_terminated():
            logger.critical("ffmpeg was terminated!")

        @ffmpeg.on("error")
        def on_error(code):
            logger.error(f"ffmpeg error code {code}")

        logger.info("Running ffmpeg")
        await ffmpeg.execute()
        if delete_original:
            os.remove(self.download.filename)

        logger.debug("Creating copy of download object")
        edited_download = copy.copy(self.download)
        edited_download.filename = Path(output)
        return edited_download



from typing import Any

from downmixer.library import Album, Artist, Song, Playlist


class SpotifyArtist(Artist):
    @classmethod
    def from_provider(
        cls, data: dict[str, Any], extra_data: dict[str, Any] = None
    ) -> "SpotifyArtist":
        return cls(
            name=data["name"],
            images=data["images"] if "images" in data.keys() else None,
            # TODO: Test the data structure of genres from Spotify
            genres=data["genres"] if "genres" in data.keys() else None,
            id=data["uri"],
            url=data["external_urls"]["spotify"],
        )


class SpotifyAlbum(Album):
    @classmethod
    def from_provider(
        cls, data: dict[str, Any], extra_data: dict[str, Any] = None
    ) -> "SpotifyAlbum":
        return cls(
            available_markets=data["available_markets"],
            name=data["name"],
            artists=SpotifyArtist.from_provider_list(data["artists"]),
            date=data["release_date"],
            track_count=data["total_tracks"],
            cover=(data["images"][0]["url"] if len(data["images"]) > 0 else None),
            id=data["uri"],
            url=data["external_urls"]["spotify"],
        )

    @classmethod
    def from_provider_list(
        cls, data: list[Any], extra_data: dict = None
    ) -> list["SpotifyAlbum"]:
        """Takes in a list of albums from the Spotify API and returns a list of SpotifyAlbums."""
        return [cls.from_provider(x["album"], extra_data) for x in data]


class SpotifySong(Song):
    @classmethod
    def from_provider(
        cls, data: dict[str, Any], extra_data: dict[str, Any] = None
    ) -> "SpotifySong":
        if "album" in data.keys():
            album = data["album"]
        elif "album" in extra_data.keys():
            album = extra_data["album"]
        else:
            album = None

        return cls(
            available_markets=data["available_markets"],
            name=data["name"],
            artists=SpotifyArtist.from_provider_list(data["artists"]),
            album=(SpotifyAlbum.from_provider(album) if album else None),
            duration=data["duration_ms"] / 1000,
            date=data["release_date"] if "release_date" in data.keys() else None,
            track_number=data["track_number"],
            isrc=(
                data["external_ids"]["isrc"] if "external_ids" in data.keys() else None
            ),
            id=data["uri"],
            url=data["external_urls"]["spotify"],
            cover=(
                album["images"][0]["url"]
                if album and len(album["images"]) > 0
                else None
            ),
        )

    @classmethod
    def from_provider_list(
        cls, data: list[dict], extra_data: dict[str, Any] = None
    ) -> list["SpotifySong"]:
        """Takes in a list of tracks from the Spotify API and returns a list of SpotifySongs."""
        try:
            return [cls.from_provider(x["track"], extra_data) for x in data]
        except KeyError:
            return [cls.from_provider(x, extra_data) for x in data]


class SpotifyPlaylist(Playlist):
    @classmethod
    def from_provider(
        cls, data: dict[str, Any], extra_data: dict[str, Any] = None
    ) -> "SpotifyPlaylist":
        return cls(
            name=data["name"],
            description=data["description"],
            tracks=SpotifySong.from_provider_list(data["tracks"]["items"]),
            images=data["images"],
            id=data["uri"],
            url=data["url"],
        )



from __future__ import annotations

import logging
from typing import Any, Optional
from urllib.request import urlopen

import mutagen

# noinspection PyProtectedMember
from mutagen.id3 import APIC, USLT, ID3

from downmixer.providers import Download

logger = logging.getLogger("downmixer").getChild(__name__)


def tag_download(download: Download):
    """Tag the Download with metadata from its `song` attribute, overriding existing metadata.

    Args:
        download (Download): Downloaded file to be tagged with song data.
    """
    logger.info(f"Tagging file {download.filename}")
    _save_easy_tag(download)

    has_cover = (
        download.song.album.cover is not None and len(download.song.album.cover) != 0
    )
    if download.song.lyrics or has_cover:
        _save_advanced_tag(download, has_cover)


def _save_easy_tag(download: Download):
    easy_id3 = mutagen.File(download.filename, easy=True)

    logger.debug("Deleting old tag information")
    easy_id3.delete()

    logger.debug(f"Filling with info from attached song '{download.song.title}'")
    easy_id3["title"] = download.song.name
    easy_id3["titlesort"] = download.song.name
    easy_id3["artist"] = download.song.all_artists
    easy_id3["isrc"] = _return_if_valid(download.song.isrc)
    easy_id3["album"] = _return_if_valid(download.song.album.name)
    easy_id3["date"] = _return_if_valid(download.song.date)
    easy_id3["originaldate"] = _return_if_valid(download.song.date)
    easy_id3["albumartist"] = _return_if_valid(download.song.album.artists)
    easy_id3["tracknumber"] = [
        _return_if_valid(download.song.track_number),
        _return_if_valid(download.song.album.track_count),
    ]
    # TODO: include all tags possible here, grab them from youtube/spotify if needed

    logger.info("Saving EasyID3 data to file")
    easy_id3.save()


def _save_advanced_tag(download: Download, has_cover):
    id3 = ID3(download.filename)
    if download.song.lyrics:
        logger.debug("Adding lyrics")
        id3["USLT::'eng'"] = USLT(
            encoding=3, lang="eng", desc="Unsynced Lyrics", text=download.song.lyrics
        )
    if has_cover:
        url = download.song.album.cover
        logger.debug(f"Downloading cover image from URL {url}")

        with urlopen(url) as raw_image:
            id3["APIC"] = APIC(
                encoding=3,
                mime="image/jpeg",
                type=3,
                desc="Cover",
                data=raw_image.read(),
            )
    logger.info("Saving ID3 data to file")
    id3.save()


def _return_if_valid(value: Any | None) -> Optional[Any]:
    if value is None:
        return ""
    else:
        return value
